import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;

public class Cop extends MazeObject {
    GameSimulator sim;
    boolean gotPlayer;
    boolean superCop;
    int eating;
    int life;
    final static int SEE_DISTANCE = 4;
    boolean chasingDoughnut;
    boolean chasingPlayer;
    Player chasedPlayer;
    Clue clue;
    Position cluePos;

    public Cop(GameSimulator sim) {
	this(0,0, sim);
	po.type = PerceptObject.COP;
	po.special = -1;
    }

    /** Makes a cop that doesn't disappear if superCop is true */
    public Cop(GameSimulator sim, boolean superCop) {	
	this(0,0, sim);
	this.superCop = superCop;
	if (superCop) {
	    this.imageFileName = "PoliceCar.gif";
	}
	po.type = PerceptObject.COP;
    }

    public Cop(int x, int y, GameSimulator sim) {
	super(x,y);
	superCop = false;
	this.sim = sim;
	gotPlayer = false;
	eating = 0;
	life = 0;
	chasingDoughnut = false;
        //this.imgName = "police";
	this.imageFileName = "Police.gif";
    }

    public void setClue(Clue c) {
	clue = c;
	cluePos = c.getPosition();
    }

    public int getEating() { return eating; }

    public void setEating(int eatTime) {
	eating = eatTime;
	if (eatTime > 0)
	    po.special = eatTime;
	else
	    po.special = -1;
    }

    public int nextMove() {
	chasingPlayer = false;
	life++;
	// 1st priority - stay with player if questioning him
	if (gotPlayer == true)
	    return -1;
	// don't bother a doughnut eating cop
	if (eating > 0) {
	    if (eating > 30)
		setEating(0);
	    else
		eating++;
	    return -1;
	}
	// cops are slow
	if (GamePlayer.nextFloat() > 0.5)
	    return -1;
	// they get bored after a while
	if (life > 180 && !superCop) {
	    delete = true;
	    return -1;
	}
	CopPercept p = getPercept();
	// 2nd priority - go to doughnut
	if (p.hasDoughnut == true) {
	    chasingDoughnut = true;
	    return moveToward(p.doughnutPos);
	}
	chasingDoughnut = false;
	// 3rd priority - apprehend player (unless already apprehended)
	if (p.hasPlayer == true) {
	    chasingPlayer = true;
	    chasedPlayer = p.player;
	    return moveToward(p.player.getPosition());
	}
	// 4th priority - snoop around clue
	if (clue != null && clue.visible == true) {
	    int distance = GameSimulator.distanceTo
		(cluePos, this.x(), this.y());
	    if (distance < 5)
		return moveNear(cluePos, distance);
	}
	// if no donut, player, or clue, all except supercop go away
	if (superCop)
	    return moveRandomly();
	else {
	    delete = true;
	    return -1;
	}
    }

    private int moveToward(Position pos) {
	if (x() == pos.x)
	    return (y() > pos.y) ? 0 : 2;
	if (y() == pos.y)
	    return (x() < pos.x) ? 1 : 3;
	return (GamePlayer.nextFloat() < 0.5) ? 
	    ((x() < pos.x) ? 1 : 3) : ((y() > pos.y) ? 0 : 2);
    }

    private int moveRandomly() {
	return (int)(Math.floor(GamePlayer.nextFloat() * 4));	
    }	    

    private int moveNear(Position pos, int distance) {
	// if in general area but far, move toward
	if (distance > 3)
	    return moveToward(pos);
	// otherwise move randomly, though away moves are restricted
	int direction;
	do {
	    direction = (int)(Math.floor(GamePlayer.nextFloat() * 4));
	}
	while (directionLegal(pos, direction, distance) == false);
	return direction;
    }

    private boolean directionLegal(Position pos, int direction, int distance) {
	if (distance == 1)
	    return true;
	Position newpos = 
	    GameSimulator.getNewPos(this.x(), this.y(), direction);
	int newdist = GameSimulator.distanceTo(pos, newpos.x, newpos.y);
	if (distance == 2) {
	    if (newdist > 2)
		return (GamePlayer.nextFloat() > 0.5) ? true : false;
	}
	else if (newdist > 3)
	    return false;
	return true;
    }

    private CopPercept getPercept() {
	CopPercept p = new CopPercept();
	Position pos = new Position(this.x(), this.y());
	for (int i=1; i<SEE_DISTANCE+1; i++) {
	    for (int j=1; j<5; j++) {
		for (int k=0; k<(2*i); k++) {
		    if (j==1 && k==0)
			pos = GameSimulator.getNewPos(pos.x, pos.y, 0);
		    else
			pos = GameSimulator.getNewPos(pos.x, pos.y, j%4);
		    if (p.hasPlayer && p.hasDoughnut)
			return p;
		    else {
			if (sim.theMaze.CoordsAreInMaze(pos.x, pos.y) == true)
			    updatePercept(p, pos);
		    }
		}
	    }
	}
	return p;
    }

    private void updatePercept(CopPercept p, Position pos) {
	MazeObject mo = sim.world[pos.x][pos.y];
	if (mo != null) {
	    if (p.hasPlayer == false && (mo instanceof Player) &&
		((Player)mo).held() == 0) {
		p.hasPlayer = true;
		p.player = (Player)mo;
		//p.playerPos = mo.getPosition();
	    }
	    else if (p.hasDoughnut == false && (mo instanceof Doughnut)) {
		p.hasDoughnut = true;
		p.doughnutPos = mo.getPosition();
	    }
	    else if (clue == null && (mo instanceof Clue))
		setClue((Clue)mo);
	}
    }
}
	
