/*
 * 1.2 version.
 */

import java.awt.*;
import java.util.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.font.*;
import java.awt.image.*;
import javax.swing.*;


/** This is the panel in which the maze is drawn */
public class MazePanel extends JPanel {
    static final int WIDTH = 500, HEIGHT = 500;
    final static int maxCharHeight = 15;
    final static int minFontSize = 6;

    final static Color bg = Color.white;
    final static Color fg = Color.black;
    final static Color red = Color.red;
    final static Color white = Color.white;

    final static BasicStroke stroke = new BasicStroke(5.0f);
    final static BasicStroke wideStroke = new BasicStroke(8.0f);

    final static float dash1[] = {10.0f};
    final static BasicStroke dashed = new BasicStroke(1.0f, 
                                                      BasicStroke.CAP_BUTT, 
                                                      BasicStroke.JOIN_MITER, 
                                                      10.0f, dash1, 0.0f);
    Shape aShape;
    GeneralPath aPath;
    
    MazeGraphics mg;

    Dimension totalSize;
    FontMetrics fontMetrics;

    
    public void init() {
        //Initialize drawing colors
        setBackground(bg);
        setForeground(fg);
    }
    /** constructor */
    public MazePanel(MazeGraphics mg) {
	super();
	setPreferredSize(new Dimension(670,670));
	this.mg = mg;
	//setDebugGraphicsOptions(DebugGraphics.LOG_OPTION);
    }
    /** update the panel for a new set of maze objects */
    public void UpdateWith(SimulationState simState) {
	mg.UpdateWith(simState, this);
    }

    /** The paint routine that paints the panel */
    public void paint(Graphics g) {
        Graphics2D g2 = (Graphics2D) g;
	this.invalidate();
	g2.setColor(Color.white);
	Dimension d = getSize();
	g2.fillRect(0,0,d.width,d.height);
	g2.setColor(Color.black);
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	g2.setStroke(stroke);
	g2.setFont(new Font("SansSerif", Font.PLAIN, 10));
	//	g2.translate(10,10);

	
	AffineTransform save = g2.getTransform(); // save it off
	float x_ratio = (float) d.width/(float)mg.width;
	float y_ratio = (float) d.height/(float)mg.height;
	g2.transform(AffineTransform.getScaleInstance(x_ratio,y_ratio));

	

	//	Rectangle2D aRect;
	//	((Rectangle2D )this.aShape).setRect(1,1,d.width-1,d.height-1);
	g2.setColor(Color.black);

	this.mg.paint(g2, this);

	g2.setTransform(save);
    }

}
