import java.util.*;
import java.io.*;

public class PlayerPercept {
    Maze theMaze;
    Player myPlayer;
    Position myPos;
    Position myPreviousPos;
    Position cluePos;
    LocalMap cluePlan;
    int time;
    int nclues;
    MazeObject[][] world;


    // Vector of each type of object...
    // Actually, these are vectors of ObjectTrackers 
    // which store the manhattan distance as well as the object itself.

    LinkedList doughnuts;
    LinkedList leaves;
    LinkedList cops;
    LinkedList bulls;
    LinkedList others;

    boolean wallUp;
    boolean wallRight;
    boolean wallDown;
    boolean wallLeft;

    boolean clearUp;
    boolean clearRight;
    boolean clearDown;
    boolean clearLeft;

    private LinkedList getRightVector(int type) {
	switch (type) {
	case PerceptObject.PLAYER:
	    return others;
	case PerceptObject.JAY:
	    return leaves;
	case PerceptObject.COP:
	    return cops;
	case PerceptObject.REDBULL:
	    return bulls;
	case PerceptObject.DOUGHNUT:
	    return doughnuts;
	default:
	    return others;
	}
    }

    /** Adds an object to the tracker lists kept in the percept */
    public void addObject(PerceptObject obj) {
	addObject(obj, getRightVector(obj.type));
    }

    /** Deletes an object from the tracker lists kept in the percept */
    public void removeObject(PerceptObject po) {
	LinkedList v = getRightVector(po.type);
	
	for (ListIterator i = v.listIterator(0); i.hasNext();) {
	    ObjectTracker ot = (ObjectTracker)(i.next());
	    if (ot.po == po) {
		if (!v.remove(ot))
		    System.out.println("Something is wrong...");
		return;
	    }
	}
	System.out.println("didn't find the object I wanted to remove...");
    }
    
    public static int mandist(int x1, int y1, int x2, int y2) {
	return Math.abs(x1 - x2) + Math.abs(y1 - y2);	
    }

    private void addObject(PerceptObject po, LinkedList v) {
	int dist =  Math.abs(po.x - myPos.x) + Math.abs(po.y - myPos.y);
	ObjectTracker ot = new ObjectTracker(po, dist);
	int k=0;
	for (ListIterator i = v.listIterator(0); i.hasNext();) {
	    ObjectTracker ot2 = (ObjectTracker)(i.next());	
	    if (ot.distance < ot2.distance)
		break;
	    k++;
	}
	v.add(k, ot);
    }

    private void sortTrackers(LinkedList v) {
	if (v.size() < 1)
	    return;
	ListIterator i;
	int j;
	for (i = v.listIterator(0),j=0; i.hasNext();j++) {
	    ObjectTracker ot = (ObjectTracker)(i.next());
	    ot.distance = Math.abs(ot.po.x - myPos.x) + 
		Math.abs(ot.po.y - myPos.y);
	}
	Collections.sort((List)v,(Comparator)v.getFirst());
    }

    public void updatePercept(Player player, GameSimulator sim) {
	// update easy positions
	myPlayer = player;
	myPreviousPos = myPos;
	myPos = player.getPosition();
	cluePos = player.getCluePosition();
	if (cluePos != null && !cluePos.valid()) {
	    cluePos = null;
	}
	time = sim.getTimeElapsed();
	nclues = player.getCluesSolved();
	cluePlan = player.getCluePlan();

	this.world = sim.world;

	// Now, must reorder the objects to be closest to the player.
	sortTrackers(doughnuts);
	sortTrackers(cops);
	sortTrackers(leaves);
	sortTrackers(bulls);
	sortTrackers(others);

	// Now, the walls and clear...
	wallUp =  (theMaze.CoordsAreInMaze(myPos.x, myPos.y-1) &&
		   theMaze.wallIsBetween(myPos.x, myPos.y, 
					 myPos.x, myPos.y-1));
	clearUp =  (!wallUp && theMaze.CoordsAreInMaze(myPos.x, myPos.y-1) &&
		    (sim.world[myPos.x][myPos.y-1] == null ||
		     ( !(sim.world[myPos.x][myPos.y-1] instanceof Player) &&
		       !(sim.world[myPos.x][myPos.y-1] instanceof Cop))));
	wallRight =  (theMaze.CoordsAreInMaze(myPos.x+1, myPos.y) &&
		      theMaze.wallIsBetween(myPos.x, myPos.y, 
					    myPos.x+1, myPos.y));
	clearRight =  (!wallRight && theMaze.CoordsAreInMaze(myPos.x+1, myPos.y) &&
		    (sim.world[myPos.x+1][myPos.y] == null ||
		     ( !(sim.world[myPos.x+1][myPos.y] instanceof Player) &&
		       !(sim.world[myPos.x+1][myPos.y] instanceof Cop))));
	wallLeft =  (theMaze.CoordsAreInMaze(myPos.x-1, myPos.y) &&
		     theMaze.wallIsBetween(myPos.x, myPos.y, 
					   myPos.x-1, myPos.y));
	clearLeft =  (!wallLeft && theMaze.CoordsAreInMaze(myPos.x-1, myPos.y) &&
		    (sim.world[myPos.x-1][myPos.y] == null ||
		     ( !(sim.world[myPos.x-1][myPos.y] instanceof Player) &&
		       !(sim.world[myPos.x-1][myPos.y] instanceof Cop))));
	wallDown =  (theMaze.CoordsAreInMaze(myPos.x, myPos.y+1) &&
		     theMaze.wallIsBetween(myPos.x, myPos.y, 
					   myPos.x, myPos.y+1));
	clearDown =  (!wallDown && theMaze.CoordsAreInMaze(myPos.x, myPos.y+1) &&
		    (sim.world[myPos.x][myPos.y+1] == null ||
		     ( !(sim.world[myPos.x][myPos.y+1] instanceof Player) &&
		       !(sim.world[myPos.x][myPos.y+1] instanceof Cop))));
	


    }

    public PlayerPercept(Player player, Maze theMaze, GameSimulator sim) {
	this.theMaze = theMaze;
	this.world = sim.world;
	myPos = player.getPosition();
	myPreviousPos = myPos;
	cluePos = null;
	int time = 0;
	int nclues = 0;
	doughnuts = new LinkedList();
	leaves = new LinkedList();
	cops = new LinkedList();
	bulls = new LinkedList();
	others = new LinkedList();
	wallUp = wallRight = wallDown = wallLeft = false;
	clearUp = clearRight = clearDown = clearLeft = false;
	cluePlan = player.getCluePlan();
    }
}
