import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

/** This is the text panel that at the moment displays the score for the 
 *  video game.  We probably want to display somethings like "# clues left", etc */
public class TextPanel extends JPanel {
    JLabel timeLabel;
    JLabel[] labels;

    public TextPanel (SimulationState simState) {
	GridBagLayout gridbag = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();
	setLayout(gridbag);
	timeLabel = new JLabel();
	c.gridx = c.gridy = 0;
	c.weightx = c.weighty = 0.1;
	c.anchor = GridBagConstraints.NORTH;
	c.ipady = 30;
	add(timeLabel, c);
	c.ipady = 10;

	labels = new JLabel[simState.numPlayers];
	for (int i=0; i<simState.numPlayers; i++) {
	    c.gridy = i+1;
	    labels[i] = new JLabel();
	    add(labels[i], c);
	}
	setPreferredSize(new Dimension(300,670));
	UpdateWith(simState);
    }

    private String getTime(int initHr, int initMin, int minElapsed) {
	int hoursElapsed = minElapsed / 60;
	int totalHr = initHr + hoursElapsed;
	int currentMin = initMin + (minElapsed % 60);
	if (currentMin >= 60) {
	    currentMin -= 60;
	    totalHr += 1;
	}
	int currentHr = totalHr % 12;
	if (currentHr == 0)
	    currentHr = 12;
	String meridian = (((totalHr / 12) % 2) == 0) ?
	    "am" : "pm";
	String sep = (currentMin < 10) ? ":0" : ":";
	return new String(currentHr + sep + currentMin + meridian);
    }

    private String addSpaces(String text, int numSpaces) {
	for (int j=0; j<numSpaces; j++)
	    text += " ";
	return text;
    }

    /** function to update the status of the text due to changes in
        the player state */
    public void UpdateWith(SimulationState simState) {
	timeLabel.setText(getTime(10, 0, 
				  simState.timeElapsed));
	if (simState.players[0].image == null)
	    return;
	for (int i=0; i<simState.numPlayers; i++) {
	    String text;
	    Player player = simState.players[i];
	    if (player.delete)
		text = "finished in " + player.finishtime / 60 + 
		    " hours and " + player.finishtime % 60 + " minutes";
	    else
		text = "Clue: " + player.cluesSolved + " Energy: " + 
		    player.energy + " Insp: " + player.inspiration + 
		    " Donuts: " + player.doughnutCount;
	    if (player.delete)
		labels[i].setForeground(new Color(0, 190, 0));
	    else if (player.held() > 0)
		labels[i].setForeground(Color.red);
	    else if (player.thinking > 0)
		labels[i].setForeground(new Color(115, 115, 255));
	    else
		labels[i].setForeground(Color.blue);
	    labels[i].setText(text);
	    labels[i].setIcon(new ImageIcon(player.image));
	    labels[i].setIconTextGap(10);
	}
    }
}
